"""
Query filter for comparison operations.
"""
from .base import QueryFilter


class ComparisonQueryFilter(QueryFilter):
    """
    Generic query filter for comparison operations (=, <, >, <=, >=, <>, !=).
    """

    def __init__(self, column: str, operator: str, value: str = None, value_placeholder: str = None):
        """
        Initialize the filter.

        This will also validate arguments.
        """
        if operator not in {"=", "<", ">", "<=", ">=", "<>", "!="}:
            raise ValueError(f"Invalid operator: {operator}")

        self.operator = operator
        self.validate_argument_exclusivity(value, value_placeholder)

        self.column = column
        self.value = value
        self.value_placeholder = value_placeholder

    def to_sql(self) -> str:
        if self.value is not None:
            return f"{self.column} {self.operator} {self.value_to_sql(self.value)}"
        else:
            return f"{self.column} {self.operator} %({self.value_placeholder})s"



# Generated migration for MariaDB UUID field conversion (Django 5.2)
"""
Migration to convert UUIDField from char(32) to uuid type for MariaDB compatibility.
"""

from django.db import migrations


def apply_mariadb_migration(apps, schema_editor):
    connection = schema_editor.connection
    if connection.vendor != 'mysql':
        return
    with connection.cursor() as cursor:
        cursor.execute("SELECT VERSION()")
        version = cursor.fetchone()[0]
        if 'mariadb' not in version.lower():
            return
    with connection.cursor() as cursor:
        # EnterpriseEnrollment
        cursor.execute("ALTER TABLE enterprise_data_enterpriseenrollment MODIFY enterprise_customer_uuid uuid NOT NULL")
        cursor.execute("ALTER TABLE enterprise_data_enterpriseenrollment MODIFY budget_id uuid NULL")
        cursor.execute("ALTER TABLE enterprise_data_enterpriseenrollment MODIFY subscription_license_uuid uuid NULL")
        cursor.execute("ALTER TABLE enterprise_data_enterpriseenrollment MODIFY enterprise_group_uuid uuid NULL")
        # EnterpriseUser
        cursor.execute("ALTER TABLE enterprise_data_enterpriseuser MODIFY enterprise_customer_uuid uuid NOT NULL")
        # EnterpriseOffer
        cursor.execute("ALTER TABLE enterprise_data_enterpriseoffer MODIFY enterprise_customer_uuid uuid NOT NULL")
        cursor.execute("ALTER TABLE enterprise_data_enterpriseoffer MODIFY enterprise_id uuid NOT NULL")
        # EnterpriseBudget
        cursor.execute("ALTER TABLE enterprise_data_enterprisebudget MODIFY enterprise_customer_uuid uuid NOT NULL")
        # EnterpriseSubsidyBudget
        cursor.execute("ALTER TABLE enterprise_data_enterprisesubsidybudget MODIFY subsidy_access_policy_uuid uuid NOT NULL")
        cursor.execute("ALTER TABLE enterprise_data_enterprisesubsidybudget MODIFY subsidy_uuid uuid NOT NULL")
        cursor.execute("ALTER TABLE enterprise_data_enterprisesubsidybudget MODIFY enterprise_customer_uuid uuid NOT NULL")
        # EnterpriseLearner
        cursor.execute("ALTER TABLE enterprise_data_enterpriselearner MODIFY enterprise_customer_id uuid NULL")
        cursor.execute("ALTER TABLE enterprise_data_enterpriselearner MODIFY enterprise_group_uuid uuid NULL")
        cursor.execute("ALTER TABLE enterprise_data_enterpriselearner MODIFY enterprise_group_membership_uuid uuid NULL")
        # EnterpriseSubsidyTransaction
        cursor.execute("ALTER TABLE enterprise_data_enterprisesubsidytransaction MODIFY enterprise_id uuid NOT NULL")
        cursor.execute("ALTER TABLE enterprise_data_enterprisesubsidytransaction MODIFY subsidy_transaction_id uuid NULL")
        cursor.execute("ALTER TABLE enterprise_data_enterprisesubsidytransaction MODIFY enterprise_customer_uuid uuid NULL")


def reverse_mariadb_migration(apps, schema_editor):
    connection = schema_editor.connection
    if connection.vendor != 'mysql':
        return
    with connection.cursor() as cursor:
        cursor.execute("SELECT VERSION()")
        version = cursor.fetchone()[0]
        if 'mariadb' not in version.lower():
            return
    with connection.cursor() as cursor:
        cursor.execute("ALTER TABLE enterprise_data_enterpriseenrollment MODIFY enterprise_customer_uuid char(32) NOT NULL")
        cursor.execute("ALTER TABLE enterprise_data_enterpriseenrollment MODIFY budget_id char(32) NULL")
        cursor.execute("ALTER TABLE enterprise_data_enterpriseenrollment MODIFY subscription_license_uuid char(32) NULL")
        cursor.execute("ALTER TABLE enterprise_data_enterpriseenrollment MODIFY enterprise_group_uuid char(32) NULL")
        cursor.execute("ALTER TABLE enterprise_data_enterpriseuser MODIFY enterprise_customer_uuid char(32) NOT NULL")
        cursor.execute("ALTER TABLE enterprise_data_enterpriseoffer MODIFY enterprise_customer_uuid char(32) NOT NULL")
        cursor.execute("ALTER TABLE enterprise_data_enterpriseoffer MODIFY enterprise_id char(32) NOT NULL")
        cursor.execute("ALTER TABLE enterprise_data_enterprisebudget MODIFY enterprise_customer_uuid char(32) NOT NULL")
        cursor.execute("ALTER TABLE enterprise_data_enterprisesubsidybudget MODIFY subsidy_access_policy_uuid char(32) NOT NULL")
        cursor.execute("ALTER TABLE enterprise_data_enterprisesubsidybudget MODIFY subsidy_uuid char(32) NOT NULL")
        cursor.execute("ALTER TABLE enterprise_data_enterprisesubsidybudget MODIFY enterprise_customer_uuid char(32) NOT NULL")
        cursor.execute("ALTER TABLE enterprise_data_enterpriselearner MODIFY enterprise_customer_id char(32) NULL")
        cursor.execute("ALTER TABLE enterprise_data_enterpriselearner MODIFY enterprise_group_uuid char(32) NULL")
        cursor.execute("ALTER TABLE enterprise_data_enterpriselearner MODIFY enterprise_group_membership_uuid char(32) NULL")
        cursor.execute("ALTER TABLE enterprise_data_enterprisesubsidytransaction MODIFY enterprise_id char(32) NOT NULL")
        cursor.execute("ALTER TABLE enterprise_data_enterprisesubsidytransaction MODIFY subsidy_transaction_id char(32) NULL")
        cursor.execute("ALTER TABLE enterprise_data_enterprisesubsidytransaction MODIFY enterprise_customer_uuid char(32) NULL")


class Migration(migrations.Migration):
    dependencies = [
        ('enterprise_data', '0048_alter_enterpriseexecedlcmoduleperformance_avg_after_lo_score_and_more'),
    ]
    operations = [
        migrations.RunPython(
            code=apply_mariadb_migration,
            reverse_code=reverse_mariadb_migration,
        ),
    ]



"""
Query Filters to filter query data.
"""
from .base import QueryFilters
from .between import BetweenQueryFilter
from .comparison import ComparisonQueryFilter
from .equal import EqualQueryFilter
from .in_ import INQueryFilter
from .null import NULLQueryFilter



# Generated migration for MariaDB UUID field conversion (Django 5.2)
"""
Migration to convert UUIDField from char(32) to uuid type for MariaDB compatibility.
"""

from django.db import migrations


def apply_mariadb_migration(apps, schema_editor):
    connection = schema_editor.connection
    if connection.vendor != 'mysql':
        return
    with connection.cursor() as cursor:
        cursor.execute("SELECT VERSION()")
        version = cursor.fetchone()[0]
        if 'mariadb' not in version.lower():
            return
    with connection.cursor() as cursor:
        cursor.execute("ALTER TABLE enterprise_data_roles_enterprisedataroleassignment MODIFY enterprise_id uuid NULL")


def reverse_mariadb_migration(apps, schema_editor):
    connection = schema_editor.connection
    if connection.vendor != 'mysql':
        return
    with connection.cursor() as cursor:
        cursor.execute("SELECT VERSION()")
        version = cursor.fetchone()[0]
        if 'mariadb' not in version.lower():
            return
    with connection.cursor() as cursor:
        cursor.execute("ALTER TABLE enterprise_data_roles_enterprisedataroleassignment MODIFY enterprise_id char(32) NULL")


class Migration(migrations.Migration):
    dependencies = [
        ('enterprise_data_roles', '0007_enterprisedataroleassignment_applies_to_all_contexts'),
    ]
    operations = [
        migrations.RunPython(
            code=apply_mariadb_migration,
            reverse_code=reverse_mariadb_migration,
        ),
    ]



"""
Module for interacting with the skills_daily_rollup_admin_dash table.
"""
from datetime import date
from typing import Optional
from uuid import UUID

from enterprise_data.admin_analytics.database.filters.mixins import CommonFiltersMixin
from enterprise_data.admin_analytics.database.queries.skills_daily_rollup_admin_dash import (
    SkillsDailyRollupAdminDashQueries,
)
from enterprise_data.admin_analytics.database.query_filters import ComparisonQueryFilter, EqualQueryFilter
from enterprise_data.admin_analytics.database.tables.base import BaseTable
from enterprise_data.admin_analytics.database.tables.fact_enrollment_admin_dash import FactEnrollmentAdminDashTable
from enterprise_data.admin_analytics.database.utils import run_query
from enterprise_data.cache.decorators import cache_it

from ..query_filters import QueryFilters


class SkillsDailyRollupAdminDashTable(CommonFiltersMixin, BaseTable):
    """
    Class for communicating with the skills_daily_rollup_admin_dash table.
    """
    queries = SkillsDailyRollupAdminDashQueries()

    def build_query_filters(
        self,
        enterprise_customer_uuid: UUID,
        start_date: date,
        end_date: date,
        course_type: Optional[str] = None,
        course_key: Optional[str] = None,
        budget_uuid: Optional[str] = None,
        include_date_range_filter: Optional[bool] = True,
    ):
        """
        Build query filters and parameters for enterprise analytics queries.

        Arguments:
            enterprise_customer_uuid (UUID): The UUID of the enterprise customer.
            start_date (date): The start date for the query.
            end_date (date): The end date for the query.
            course_type (str, optional): The course type to filter by (e.g., 'OCM', 'Executive Education').
            course_key (str, optional): The course key to filter by.
            budget_uuid (str, optional): The budget UUID to filter by.
            include_date_range_filter (bool, optional): Whether to include the date range filter. Defaults to True.

        Returns:
            tuple: A tuple containing:
                - QueryFilters: The filters to apply to the query.
                - dict: The parameters to use in the query.
        """
        optional_params = {}
        default_params = {
            'enterprise_customer_uuid': enterprise_customer_uuid,
        }

        query_filters = QueryFilters([
            self.enterprise_customer_uuid_filter('enterprise_customer_uuid'),
        ])

        if include_date_range_filter is True:
            query_filters.append(
                self.date_range_filter(
                    column='date',
                    start_date_params_key='start_date',
                    end_date_params_key='end_date',
                )
            )
            default_params.update({
                'start_date': start_date,
                'end_date': end_date,
            })

        if course_key:
            query_filters.append(EqualQueryFilter(
                column='course_key',
                value_placeholder='course_key',
            ))
            optional_params['course_key'] = course_key

        if course_type:
            query_filters.append(EqualQueryFilter(
                column='course_product_line',
                value_placeholder='course_type',
            ))
            optional_params['course_type'] = course_type

        if budget_uuid:
            query_filters.append(EqualQueryFilter(
                column='subsidy_access_policy_uuid',
                value_placeholder='budget_uuid',
            ))
            optional_params['budget_uuid'] = budget_uuid

        params = {**default_params, **optional_params}
        return query_filters, params

    @cache_it()
    def get_top_skills(
        self,
        enterprise_customer_uuid: UUID,
        start_date: date,
        end_date: date,
        course_type: Optional[str] = None,
        course_key: Optional[str] = None,
        budget_uuid: Optional[str] = None,
    ):
        """
        Get the top skills for the given enterprise customer.

        Arguments:
            enterprise_customer_uuid (UUID): The UUID of the enterprise customer.
            start_date (date): The start date.
            end_date (date): The end date.
            course_type (str): The course type (OCM or Executive Education) to filter by (optional).
            course_key (str): The course key to filter by (optional).
            budget_uuid (str): The budget UUID to filter by (optional).

        Returns:
            list<dict>: A list of dictionaries containing the skill_name, skill_type, enrolls and completions.
        """
        query_filters, params = self.build_query_filters(
            enterprise_customer_uuid=enterprise_customer_uuid,
            start_date=start_date,
            end_date=end_date,
            course_type=course_type,
            course_key=course_key,
            budget_uuid=budget_uuid,
        )
        return run_query(
            query=self.queries.get_top_skills(query_filters),
            params=params,
            as_dict=True,
        )

    @cache_it()
    def get_top_skills_by_enrollment(
        self,
        enterprise_customer_uuid: UUID,
        start_date: date,
        end_date: date,
        course_type: Optional[str] = None,
        course_key: Optional[str] = None,
        budget_uuid: Optional[str] = None,
    ):
        """
        Get the top skills by enrollments for the given enterprise customer.

        Arguments:
            enterprise_customer_uuid (UUID): The UUID of the enterprise customer.
            start_date (date): The start date.
            end_date (date): The end date.
            course_type (str): The course type (OCM or Executive Education) to filter by (optional).
            course_key (str): The course key to filter by (optional).
            budget_uuid (str): The budget UUID to filter by (optional).

        Returns:
            list<dict>: A list of dictionaries containing the skill_name, subject_name, count.
        """
        query_filters, params = self.build_query_filters(
            enterprise_customer_uuid=enterprise_customer_uuid,
            start_date=start_date,
            end_date=end_date,
            course_type=course_type,
            course_key=course_key,
            budget_uuid=budget_uuid,
        )
        return run_query(
            query=self.queries.get_top_skills_by_enrollment(query_filters),
            params=params,
            as_dict=True,
        )

    @cache_it()
    def get_top_skills_by_completion(
        self,
        enterprise_customer_uuid: UUID,
        start_date: date,
        end_date: date,
        course_type: Optional[str] = None,
        course_key: Optional[str] = None,
        budget_uuid: Optional[str] = None,
    ):
        """
        Get the top skills by completion for the given enterprise customer.

        Arguments:
            enterprise_customer_uuid (UUID): The UUID of the enterprise customer.
            start_date (date): The start date.
            end_date (date): The end date.
            course_type (str): The course type (OCM or Executive Education) to filter by (optional).
            course_key (str): The course key to filter by (optional).
            budget_uuid (str): The budget UUID to filter by (optional).

        Returns:
            list<dict>: A list of dictionaries containing the skill_name, subject_name, count.
        """
        query_filters, params = self.build_query_filters(
            enterprise_customer_uuid=enterprise_customer_uuid,
            start_date=start_date,
            end_date=end_date,
            course_type=course_type,
            course_key=course_key,
            budget_uuid=budget_uuid,
        )
        return run_query(
            query=self.queries.get_top_skills_by_completion(query_filters),
            params=params,
            as_dict=True,
        )

    @cache_it()
    def get_skills_by_learning_hours(
        self,
        enterprise_customer_uuid: UUID,
        start_date: date,
        end_date: date,
        course_key: Optional[str] = None,
        course_type: Optional[str] = None,
        budget_uuid: Optional[str] = None,
    ):
        """
        Get the skills by learning hours for the given enterprise customer.

        Arguments:
            enterprise_customer_uuid (UUID): The UUID of the enterprise customer.
            start_date (date): The start date.
            end_date (date): The end date.
            course_key (str): The course key to filter by (optional).
            course_type (str): The course type (OCM or Executive Education) to filter by (optional).
            budget_uuid (str): The budget UUID to filter by (optional).

        Returns:
            list<dict>: A list of dictionaries containing the skill_name, subject_name, count.
        """
        query_filters, params = self.build_query_filters(
            enterprise_customer_uuid=enterprise_customer_uuid,
            start_date=start_date,
            end_date=end_date,
            course_key=course_key,
            course_type=course_type,
            budget_uuid=budget_uuid,
        )
        return run_query(
            query=self.queries.get_skills_by_learning_hours(query_filters),
            params=params,
            as_dict=True,
        )

    @cache_it()
    def get_unique_skills_gained(
        self,
        enterprise_customer_uuid: UUID,
        start_date: date,
        end_date: date,
        course_type: Optional[str] = None,
        course_key: Optional[str] = None,
        budget_uuid: Optional[str] = None,
    ):
        """
        Get the unique skills gained for the given enterprise customer.

        Args:
            enterprise_customer_uuid (UUID): The UUID of the enterprise customer.
            start_date (date): The start date.
            end_date (date): The end date.
            course_type (str): The course type (OCM or Executive Education) to filter by (optional).
            course_key (str): The course key to filter by (optional). Defaults to None.
            budget_uuid (str): The budget UUID to filter by (optional). Defaults to None.
        """
        query_filters, params = self.build_query_filters(
            enterprise_customer_uuid=enterprise_customer_uuid,
            start_date=start_date,
            end_date=end_date,
            course_key=course_key,
            course_type=course_type,
            budget_uuid=budget_uuid,
        )

        query_filters.append(ComparisonQueryFilter(
            column='completions',
            operator='>',
            value=0
        ))

        results = run_query(
            query=self.queries.get_unique_skills_gained(query_filters),
            params=params,
        )

        return results[0][0] if results else 0

    def construct_upskill_learners_query_filters(
        self,
        enterprise_customer_uuid: UUID,
        start_date: date,
        end_date: date,
        course_type: Optional[str] = None,
        course_key: Optional[str] = None,
        budget_uuid: Optional[str] = None,
    ):
        """
        Construct query filters and parameters for upskilled learners query.

        Args:
            enterprise_customer_uuid (UUID): The UUID of the enterprise customer.
            start_date (date): The start date.
            end_date (date): The end date.
            course_type (str): The course type (OCM or Executive Education) to filter by (optional).
            course_key (str): The course key to filter by (optional). Defaults to None.
            budget_uuid (str): The budget UUID to filter by (optional). Defaults to None.
        """
        skills_query_filters, skills_params = self.build_query_filters(
            enterprise_customer_uuid=enterprise_customer_uuid,
            start_date=start_date,
            end_date=end_date,
            course_key=course_key,
            course_type=course_type,
            budget_uuid=budget_uuid,
        )
        skills_query_filters.append(ComparisonQueryFilter(
            column='completions',
            operator='>',
            value=0
        ))

        enrollment_query_filters, enrollment_params = FactEnrollmentAdminDashTable().build_query_filters(
            enterprise_customer_uuid=enterprise_customer_uuid,
            start_date=start_date,
            end_date=end_date,
            course_key=course_key,
            course_type=course_type,
            budget_uuid=budget_uuid,
        )
        enrollment_query_filters.append(ComparisonQueryFilter(
            column='has_passed',
            operator='=',
            value=1
        ))

        return skills_query_filters, skills_params, enrollment_query_filters, enrollment_params

    @cache_it()
    def get_upskilled_learners_count(
        self,
        enterprise_customer_uuid: UUID,
        start_date: date,
        end_date: date,
        course_type: Optional[str] = None,
        course_key: Optional[str] = None,
        budget_uuid: Optional[str] = None,
    ):
        """
        Get the count of upskilled learners for the given enterprise customer.

        Args:
            enterprise_customer_uuid (UUID): The UUID of the enterprise customer.
            start_date (date): The start date.
            end_date (date): The end date.
            course_type (str): The course type (OCM or Executive Education) to filter by (optional).
            course_key (str): The course key to filter by (optional). Defaults to None.
            budget_uuid (str): The budget UUID to filter by (optional). Defaults to None.
        """
        skills_filters, skills_params, enroll_filters, enroll_params = self.construct_upskill_learners_query_filters(
            enterprise_customer_uuid=enterprise_customer_uuid,
            start_date=start_date,
            end_date=end_date,
            course_key=course_key,
            course_type=course_type,
            budget_uuid=budget_uuid,
        )

        results = run_query(
            query=self.queries.get_upskilled_learners_count(skills_filters, enroll_filters),
            params={**skills_params, **enroll_params},
        )

        return results[0][0] if results else 0

    def construct_new_skills_learned_query_filters(
        self,
        enterprise_customer_uuid: UUID,
        start_date: date,
        end_date: date,
        course_type: Optional[str] = None,
        course_key: Optional[str] = None,
        budget_uuid: Optional[str] = None,
    ):
        """
        Construct query filters and parameters for new skills learned query.

        Args:
            enterprise_customer_uuid (UUID): The UUID of the enterprise customer.
            start_date (date): The start date.
            end_date (date): The end date.
            course_type (str): The course type (OCM or Executive Education) to filter by (optional).
            course_key (str): The course key to filter by (optional). Defaults to None.
            budget_uuid (str): The budget UUID to filter by (optional). Defaults to None.
        """
        common_query_filters = QueryFilters([
            ComparisonQueryFilter(
                column='completions',
                operator='>',
                value=0
            ),
            ComparisonQueryFilter(
                column='confidence',
                operator='>=',
                value=0.8
            )
        ])

        current_skills_filters, current_skills_params = self.build_query_filters(
            enterprise_customer_uuid=enterprise_customer_uuid,
            start_date=start_date,
            end_date=end_date,
            course_key=course_key,
            course_type=course_type,
            budget_uuid=budget_uuid,
        )
        current_skills_filters.extend(common_query_filters)

        historical_skills_filters, hist_skills_params = self.build_query_filters(
            enterprise_customer_uuid=enterprise_customer_uuid,
            start_date=start_date,
            end_date=end_date,
            course_key=course_key,
            course_type=course_type,
            budget_uuid=budget_uuid,
            include_date_range_filter=False,
        )
        historical_skills_filters.append(
            ComparisonQueryFilter(
                column='date',
                operator='<',
                value_placeholder='start_date'
            )
        )
        historical_skills_filters.extend(common_query_filters)

        return current_skills_filters, current_skills_params, historical_skills_filters, hist_skills_params

    @cache_it()
    def get_new_skills_learned_count(
        self,
        enterprise_customer_uuid: UUID,
        start_date: date,
        end_date: date,
        course_type: Optional[str] = None,
        course_key: Optional[str] = None,
        budget_uuid: Optional[str] = None,
    ):
        """
        Get the count of new skills learned for the given enterprise customer.

        Args:
            enterprise_customer_uuid (UUID): The UUID of the enterprise customer.
            start_date (date): The start date.
            end_date (date): The end date.
            course_type (str): The course type (OCM or Executive Education) to filter by (optional).
            course_key (str): The course key to filter by (optional). Defaults to None.
            budget_uuid (str): The budget UUID to filter by (optional). Defaults to None.
        """
        current_filters, current_params, hist_filters, hist_params = self.construct_new_skills_learned_query_filters(
            enterprise_customer_uuid=enterprise_customer_uuid,
            start_date=start_date,
            end_date=end_date,
            course_key=course_key,
            course_type=course_type,
            budget_uuid=budget_uuid,
        )

        results = run_query(
            query=self.queries.get_new_skills_learned_count(hist_filters, current_filters),
            params={**current_params, **hist_params},
        )

        return results[0][0] if results else 0



"""
Query filter for to check if a column is NULL or NOT NULL.
"""
from .base import QueryFilter


class NULLQueryFilter(QueryFilter):
    """
    Query filter for checking if a column is NULL or NOT NULL.
    """
    def __init__(self, column: str, null_check: bool):
        """
        Initialize the filter.
        """
        self.column = column
        self.null_check = null_check

    def to_sql(self) -> str:
        return f'{self.column} IS NULL' if self.null_check else f'{self.column} IS NOT NULL'



"""
Serializers for enterprise api v1.
"""
from uuid import UUID

from rest_framework import serializers

from enterprise_data.admin_analytics.constants import CourseType, ResponseType
from enterprise_data.cache.decorators import cache_it
from enterprise_data.models import (
    EnterpriseAdminLearnerProgress,
    EnterpriseAdminSummarizeInsights,
    EnterpriseExecEdLCModulePerformance,
    EnterpriseGroupMembership,
    EnterpriseLearner,
    EnterpriseLearnerEnrollment,
    EnterpriseOffer,
    EnterpriseSubsidyBudget,
)
from enterprise_data.utils import calculate_percentage_difference


class EnterpriseLearnerEnrollmentSerializer(serializers.ModelSerializer):
    """
    Serializer for EnterpriseLearnerEnrollment model.
    """
    course_api_url = serializers.SerializerMethodField()
    enterprise_user_id = serializers.SerializerMethodField()
    total_learning_time_hours = serializers.SerializerMethodField()
    enterprise_flex_group_name = serializers.SerializerMethodField()
    enterprise_flex_group_uuid = serializers.SerializerMethodField()

    class Meta:
        model = EnterpriseLearnerEnrollment
        # Do not change the order of fields below. Ordering is important becuase `progress_v3`
        # csv generated in `enterprise_reporting` should be same as csv generated on `admin-portal`
        # Order and field names below should match with `EnrollmentsCSVRenderer.header`
        fields = (
            'enrollment_id', 'enterprise_enrollment_id', 'is_consent_granted', 'paid_by',
            'user_current_enrollment_mode', 'enrollment_date', 'unenrollment_date',
            'unenrollment_end_within_date', 'is_refunded', 'seat_delivery_method',
            'offer_id', 'offer_name', 'offer_type', 'coupon_code', 'coupon_name', 'contract_id',
            'course_list_price', 'amount_learner_paid', 'course_key', 'courserun_key',
            'course_title', 'course_pacing_type', 'course_start_date', 'course_end_date',
            'course_duration_weeks', 'course_max_effort', 'course_min_effort',
            'course_primary_program', 'primary_program_type', 'course_primary_subject', 'has_passed',
            'last_activity_date', 'progress_status', 'passed_date', 'current_grade',
            'letter_grade', 'enterprise_user_id', 'user_email', 'user_account_creation_date',
            'user_country_code', 'user_username', 'user_first_name', 'user_last_name', 'enterprise_name',
            'enterprise_customer_uuid', 'enterprise_sso_uid', 'created', 'course_api_url',
            'total_learning_time_hours', 'is_subsidy', 'course_product_line', 'budget_id',
            'enterprise_flex_group_name', 'enterprise_flex_group_uuid',
        )

    def get_course_api_url(self, obj):
        """Constructs course api url"""
        return '/enterprise/v1/enterprise-catalogs/{enterprise_customer_uuid}/courses/{courserun_key}'.format(
            enterprise_customer_uuid=obj.enterprise_customer_uuid, courserun_key=obj.courserun_key
        )

    def get_enterprise_user_id(self, obj):
        """Returns enterprise user id of a learner's enrollment"""
        return obj.enterprise_user_id

    def get_total_learning_time_hours(self, obj):
        """Returns the learners total learning time in hours"""
        return round((obj.total_learning_time_seconds or 0.0)/3600.0, 2)

    @cache_it()
    def _get_flex_groups(self, obj):
        """
        Returns list of tuples containing group (name, uuid) pairs for the learner.
        This is cached to prevent duplicate database queries.
        """
        enterprise_user_id = obj.enterprise_user_id

        if not enterprise_user_id:
            return []

        # Get all group memberships for this user in a single query
        # Order by name for consistent ordering
        return list(
            EnterpriseGroupMembership.objects.filter(
                enterprise_customer_user_id=enterprise_user_id,
                membership_is_removed=False,
                group_is_removed=False,
                group_type="flex",
            )
            .order_by("enterprise_group_name")
            .values_list("enterprise_group_name", "enterprise_group_uuid")
            .distinct()
        )

    def get_enterprise_flex_group_name(self, obj):
        """Returns a comma-separated list of enterprise group names that the learner is associated with"""
        groups = self._get_flex_groups(obj)

        if not groups:
            return obj.enterprise_group_name

        # Return comma-separated list of group names (first element of each tuple)
        return ', '.join(group[0] for group in groups)

    def get_enterprise_flex_group_uuid(self, obj):
        """Returns a comma-separated list of enterprise group UUIDs that the learner is associated with"""
        groups = self._get_flex_groups(obj)

        if not groups:
            return obj.enterprise_group_uuid

        # Return comma-separated list of group UUIDs (second element of each tuple)
        return ', '.join(str(group[1]) for group in groups)


class EnterpriseSubsidyBudgetSerializer(serializers.ModelSerializer):
    """
    Serializer for EnterpriseSubsidyBudget model.
    """

    class Meta:
        model = EnterpriseSubsidyBudget
        exclude = ('id',)


class EnterpriseOfferSerializer(serializers.ModelSerializer):
    """
    Serializer for EnterpriseOfferSerializer model.
    """
    budgets = serializers.SerializerMethodField()

    def get_budgets(self, instance):
        """Gets related budgets"""

        try:
            # if offer_id is a uuid
            offer_uuid = UUID(instance.offer_id, version=4)
            budget_queryset = EnterpriseSubsidyBudget.objects.filter(subsidy_uuid=offer_uuid)
            serializer = EnterpriseSubsidyBudgetSerializer(instance=budget_queryset, many=True, read_only=True)
            return serializer.data
        except ValueError:
            pass

        return []

    class Meta:
        model = EnterpriseOffer
        fields = '__all__'

    def to_internal_value(self, data):
        """
        Convert the incoming data offer_id field to a format that can be stored in the db.

        For a given offer_id string from the requester, determine the best representation to use for db storage.

        Raises serializers.ValidationError:
            If the given string is not exclusively numeric characters, but also does not parse as a UUID (either because
            it has the wrong length, incorrect dashes, or some other reason).
        """
        ret = super().to_internal_value(data)
        if ret['offer_id'] is None or ret['offer_id'] == '':
            raise serializers.ValidationError("requested offer_id is None.")

        if isinstance(ret['offer_id'], str) and len(ret['offer_id']) == 36:
            offer_id = ret['offer_id'].replace('-', '')
            # There should only be 4 dashes in the UUID, making the length 32 after removal
            if len(offer_id) == 32:
                ret['offer_id'] = offer_id
                return ret

            else:
                raise serializers.ValidationError("requested offer_id neither a valid integer nor UUID.")

        if len(ret['offer_id']) < 10:  # All ecommerce offer_ids are at < 1 million.
            try:
                int(ret['offer_id'])
                return ret
            except ValueError as e:
                raise serializers.ValidationError("Requested offer_id not a valid integer.") from e

        raise serializers.ValidationError("requested offer_id neither a valid integer nor UUID.")

    def to_representation(self, instance):
        """
        Add `-` dashes to the outgoing data offer_id field.
        """
        ret = super().to_representation(instance)

        # A 32 character offer_id is our heuristic for whether the stored value represents a UUID or integer.  If the
        # heuristic passes, make the serialized output look like a UUID.
        if len(ret['offer_id']) == 32:
            ret['offer_id'] = '-'.join([
                    ret['offer_id'][:8],
                    ret['offer_id'][8:12],
                    ret['offer_id'][12:16],
                    ret['offer_id'][16:20],
                    ret['offer_id'][20:]
                ]
            )

        return ret


class EnterpriseLearnerSerializer(serializers.ModelSerializer):
    """
    Serializer for EnterpriseLearner model.
    """

    class Meta:
        model = EnterpriseLearner
        fields = '__all__'

    def to_representation(self, instance):
        representation = super().to_representation(instance)

        if hasattr(instance, 'enrollment_count'):
            representation['enrollment_count'] = instance.enrollment_count
        if hasattr(instance, 'course_completion_count'):
            representation['course_completion_count'] = instance.course_completion_count

        return representation


class LearnerCompletedCoursesSerializer(serializers.Serializer):    # pylint: disable=abstract-method
    """
    Serializer for learner's completed courses.
    """
    class Meta:
        ref_name = 'v1.LearnerCompletedCoursesSerializer'

    user_email = serializers.EmailField()
    completed_courses = serializers.IntegerField()


class EnterpriseAdminLearnerProgressSerializer(serializers.ModelSerializer):
    """
    Serializer for EnterpriseAdminLearnerProgress model.
    """

    class Meta:
        model = EnterpriseAdminLearnerProgress
        fields = '__all__'


class EnterpriseAdminSummarizeInsightsSerializer(serializers.ModelSerializer):
    """
    Serializer for EnterpriseAdminSummarizeInsights model.
    """

    class Meta:
        model = EnterpriseAdminSummarizeInsights
        fields = '__all__'


class AdminAnalyticsAggregatesQueryParamsSerializer(serializers.Serializer):  # pylint: disable=abstract-method
    """
    Serializer for validating admin analytics query params.
    """
    start_date = serializers.DateField(required=False)
    end_date = serializers.DateField(required=False)
    granularity = serializers.CharField(required=False)
    calculation = serializers.CharField(required=False)
    response_type = serializers.CharField(required=False)
    page = serializers.IntegerField(required=False)
    chart_type = serializers.CharField(required=False)
    group_uuid = serializers.UUIDField(required=False)

    def validate(self, attrs):
        """
        Validate the query params.

        Raises:
            serializers.ValidationError: If start_date is greater than end_date.
        """
        if 'start_date' in attrs and 'end_date' in attrs:
            if attrs['start_date'] > attrs['end_date']:
                raise serializers.ValidationError("start_date should be less than or equal to end_date.")
        return attrs


class EnterpriseExecEdLCModulePerformanceSerializer(serializers.ModelSerializer):
    """
    Serializer for EnterpriseExecEdLCModulePerformance model.
    """
    extensions_requested = serializers.SerializerMethodField()
    avg_lo_percentage_difference = serializers.SerializerMethodField()

    class Meta:
        model = EnterpriseExecEdLCModulePerformance
        fields = '__all__'

    def get_extensions_requested(self, obj):
        """Return extensions_requested if not None, otherwise return 0"""
        return obj.extensions_requested if obj.extensions_requested is not None else 0

    def get_avg_lo_percentage_difference(self, obj):
        """
        Return percentage difference between `avg_before_lo_score` and `avg_after_lo_score` if not None,
        otherwise return None
        """
        if obj.avg_before_lo_score is None or obj.avg_after_lo_score is None:
            return None
        return round(
            calculate_percentage_difference(obj.avg_before_lo_score, obj.avg_after_lo_score),
            2
        )


class EnterpriseBudgetSerializer(serializers.ModelSerializer):
    """
    Serializer for EnterpriseSubsidyBudget model.
    """

    class Meta:
        model = EnterpriseSubsidyBudget
        fields = (
            'subsidy_access_policy_uuid',
            'subsidy_access_policy_display_name',
        )


class EnterpriseGroupMembershipSerializer(serializers.ModelSerializer):
    """
    Serializer for EnterpriseGroupMembership model.
    """

    class Meta:
        model = EnterpriseGroupMembership
        fields = (
            'enterprise_group_uuid',
            'enterprise_group_name',
        )


class AdvanceAnalyticsQueryParamSerializer(serializers.Serializer):  # pylint: disable=abstract-method
    """Serializer for validating query params"""
    RESPONSE_TYPES = [
        ResponseType.JSON.value,
        ResponseType.CSV.value
    ]
    start_date = serializers.DateField(required=False)
    end_date = serializers.DateField(required=False)
    response_type = serializers.CharField(required=False)
    page = serializers.IntegerField(required=False, min_value=1)
    page_size = serializers.IntegerField(required=False, min_value=2)
    group_uuid = serializers.UUIDField(required=False)
    course_type = serializers.ChoiceField(
        choices=[course_type.value for course_type in CourseType],
        required=False,
        allow_blank=False,
        allow_null=False,
    )
    course_key = serializers.CharField(required=False)
    budget_uuid = serializers.UUIDField(required=False, format='hex')

    def validate(self, attrs):
        """
        Validate the query params.

        Raises:
            serializers.ValidationError: If start_date is greater than end_date.
        """
        start_date = attrs.get('start_date')
        end_date = attrs.get('end_date')

        if start_date and end_date and start_date > end_date:
            raise serializers.ValidationError("start_date should be less than or equal to end_date.")

        return attrs

    def validate_response_type(self, value):
        """
        Validate the response_type value.

        Raises:
            serializers.ValidationError: If response_type is not one of the valid choices in `RESPONSE_TYPES`.
        """
        if value not in self.RESPONSE_TYPES:
            raise serializers.ValidationError(f"response_type must be one of {self.RESPONSE_TYPES}")
        return value



"""
Module for interacting with the fact_enrollment_engagement_day_admin_dash table.
"""
from datetime import date
from typing import Optional, Tuple
from uuid import UUID

from enterprise_data.admin_analytics.database.query_filters import EqualQueryFilter, INQueryFilter, NULLQueryFilter
from enterprise_data.cache.decorators import cache_it
from enterprise_data.utils import find_first

from ..filters import FactEngagementAdminDashFilters
from ..queries import FactEngagementAdminDashQueries
from ..query_filters import QueryFilters
from ..utils import run_query
from .base import BaseTable

NULL_EMAIL_TEXT = 'learners who have not shared consent'


class FactEngagementAdminDashTable(BaseTable):
    """
    Class for communicating with the fact_enrollment_engagement_day_admin_dash table.
    """
    queries = FactEngagementAdminDashQueries()
    engagement_filters = FactEngagementAdminDashFilters()

    def __get_common_query_filters_for_engagement(
            self, enterprise_customer_uuid: UUID,
            group_uuid: Optional[UUID],
            start_date: date,
            end_date: date,
            course_type: Optional[str] = None,
            course_key: Optional[str] = None,
            budget_uuid: Optional[str] = None,
    ) -> Tuple[QueryFilters, dict]:
        """
        Utility method to get query filters common in most usages below.
        This will return a tuple containing the query filters list and the dictionary of query parameters that
        will be used in the query.
        It will contain the following query filters.
            1. enterprise_customer_uuid filter to filter records for an enterprise customer.
            2. enrollment_date range filter to filter records by enrollment date.
            3. group_uuid filter to filter records for learners who belong to the given group.
        """
        query_filters = QueryFilters([
            self.engagement_filters.enterprise_customer_uuid_filter('enterprise_customer_uuid'),
            self.engagement_filters.date_range_filter('activity_date', 'start_date', 'end_date'),
        ])
        params = {
            'enterprise_customer_uuid': enterprise_customer_uuid,
            'start_date': start_date,
            'end_date': end_date,
        }

        response = self.engagement_filters.enterprise_user_query_filter(
            group_uuid,
            enterprise_customer_uuid
        )
        if response is not None:
            enterprise_user_id_in_filter, enterprise_user_id_params = response
            query_filters.append(enterprise_user_id_in_filter)
            params.update(enterprise_user_id_params)

        # add optional filters
        if course_type:
            query_filters.append(EqualQueryFilter(
                column='course_product_line',
                value_placeholder='course_type',
            ))
            params['course_type'] = course_type

        if course_key:
            query_filters.append(EqualQueryFilter(
                column='course_key',
                value_placeholder='course_key',
            ))
            params['course_key'] = course_key

        if budget_uuid:
            query_filters.append(EqualQueryFilter(
                column='subsidy_access_policy_uuid',
                value_placeholder='budget_uuid',
            ))
            params['budget_uuid'] = budget_uuid

        return query_filters, params

    @cache_it()
    def get_learning_hours_and_daily_sessions(
        self,
        enterprise_customer_uuid: UUID,
        start_date: date,
        end_date: date,
        group_uuid: Optional[UUID] = None,
        course_type: Optional[str] = None,
        course_key: Optional[str] = None,
        budget_uuid: Optional[str] = None
    ):
        """
        Get the learning hours and daily sessions for the given enterprise customer.

        Arguments:
            enterprise_customer_uuid (UUID): The UUID of the enterprise customer.
            start_date (date): The start date.
            end_date (date): The end date.
            group_uuid (UUID): The UUID of a group.
            course_type (str): The type of the course.
            course_key (str): The key of the course.
            budget_uuid (str): The UUID of the budget.

        Returns:
            (tuple<float, int>): The learning hours and daily sessions.
        """
        query_filters, query_filter_params = self.__get_common_query_filters_for_engagement(
            enterprise_customer_uuid, group_uuid, start_date, end_date, course_type, course_key, budget_uuid
        )
        results = run_query(
            query=self.queries.get_learning_hours_and_daily_sessions_query(query_filters),
            params={
                **query_filter_params,
            }
        )
        if not results:
            return 0.0, 0

        return tuple(results[0])

    @cache_it()
    def get_engagement_count(
        self,
        enterprise_customer_uuid: UUID,
        group_uuid: Optional[UUID],
        start_date: date,
        end_date: date
    ):
        """
        Get the total number of engagements for the given enterprise customer.

        Arguments:
            enterprise_customer_uuid (UUID): The UUID of the enterprise customer.
            start_date (date): The start date.
            end_date (date): The end date.

        Returns:
            int: The total number of engagements.
        """
        query_filters, query_filter_params = self.__get_common_query_filters_for_engagement(
            enterprise_customer_uuid, group_uuid, start_date, end_date
        )

        results = run_query(
            query=self.queries.get_engagement_count_query(query_filters),
            params=query_filter_params,
        )
        if not results:
            return 0
        return results[0][0]

    @cache_it()
    def get_all_engagements(
        self,
        enterprise_customer_uuid: UUID,
        group_uuid: Optional[UUID],
        start_date: date,
        end_date: date,
        limit: int,
        offset: int
    ):
        """
        Get all engagement data for the given enterprise customer.

        Arguments:
            enterprise_customer_uuid (UUID): The UUID of the enterprise customer.
            group_uuid (UUID): The UUID of a group.
            start_date (date): The start date.
            end_date (date): The end date.
            limit (int): The maximum number of records to return.
            offset (int): The number of records to skip.

        Returns:
            list<dict>: A list of dictionaries containing the engagement data.
        """
        query_filters, query_filter_params = self.__get_common_query_filters_for_engagement(
            enterprise_customer_uuid, group_uuid, start_date, end_date
        )

        return run_query(
            query=self.queries.get_all_engagement_query(query_filters),
            params={
                **query_filter_params,
                'limit': limit,
                'offset': offset,
            },
            as_dict=True,
        )

    @cache_it()
    def get_top_courses_by_engagement(
        self,
        enterprise_customer_uuid: UUID,
        group_uuid: Optional[UUID],
        start_date: date,
        end_date: date,
        course_type: Optional[str] = None,
        course_key: Optional[str] = None,
        budget_uuid: Optional[str] = None
    ):
        """
        Get the top courses by user engagement for the given enterprise customer.

        Arguments:
            enterprise_customer_uuid (UUID): The UUID of the enterprise customer.
            start_date (date): The start date.
            end_date (date): The end date.
            course_type (Optional[str]): The course type (OCM or Executive Education) to filter by (optional).
            course_key (Optional[str]): The course key to filter by (optional).
            budget_uuid (Optional[str]): The budget UUID to filter by (optional).

        Returns:
            list<dict>: A list of dictionaries containing the course data.
        """
        query_filters, query_filter_params = self.__get_common_query_filters_for_engagement(
            enterprise_customer_uuid, group_uuid, start_date, end_date, course_type, course_key, budget_uuid
        )
        return run_query(
            query=self.queries.get_top_courses_by_engagement_query(query_filters),
            params=query_filter_params,
            as_dict=True,
        )

    @cache_it()
    def get_top_subjects_by_engagement(
        self,
        enterprise_customer_uuid: UUID,
        group_uuid: Optional[UUID],
        start_date: date,
        end_date: date,
        course_type: Optional[str] = None,
        course_key: Optional[str] = None,
        budget_uuid: Optional[str] = None
    ):
        """
        Get the top subjects by user engagement for the given enterprise customer.

        Arguments:
            enterprise_customer_uuid (UUID): The UUID of the enterprise customer.
            start_date (date): The start date.
            end_date (date): The end date.
            course_type (Optional[str]): The course type (OCM or Executive Education) to filter by (optional).
            course_key (Optional[str]): The course key to filter by (optional).
            budget_uuid (Optional[str]): The budget UUID to filter by (optional).

        Returns:
            list<dict>: A list of dictionaries containing the subject data.
        """
        query_filters, query_filter_params = self.__get_common_query_filters_for_engagement(
            enterprise_customer_uuid, group_uuid, start_date, end_date, course_type, course_key, budget_uuid
        )
        return run_query(
            query=self.queries.get_top_subjects_by_engagement_query(query_filters),
            params=query_filter_params,
            as_dict=True,
        )

    @cache_it()
    def get_engagement_time_series_data(
        self,
        enterprise_customer_uuid: UUID,
        group_uuid: Optional[UUID],
        start_date: date,
        end_date: date,
        course_type: Optional[str] = None,
        course_key: Optional[str] = None,
        budget_uuid: Optional[str] = None,
    ):
        """
        Get the engagement time series data.

        Arguments:
            enterprise_customer_uuid (UUID): The UUID of the enterprise customer.
            start_date (date): The start date.
            end_date (date): The end date.
            course_type (Optional[str]): The course type (OCM or Executive Education) to filter by (optional).
            course_key (Optional[str]): The course key to filter by (optional).
            budget_uuid (Optional[str]): The budget UUID to filter by (optional).

        Returns:
            list<dict>: A list of dictionaries containing the engagement time series data.
        """
        query_filters, query_filter_params = self.__get_common_query_filters_for_engagement(
            enterprise_customer_uuid, group_uuid, start_date, end_date, course_type, course_key, budget_uuid
        )
        return run_query(
            query=self.queries.get_engagement_time_series_data_query(query_filters),
            params=query_filter_params,
            as_dict=True,
        )

    def build_query_filters_for_leaderboard(
        self,
        enterprise_customer_uuid: UUID,
        date_column: str,
        start_date: date,
        end_date: date,
        equality_filters: Optional[dict] = None,
        null_filters: Optional[list] = None,
        in_filters: Optional[dict] = None
    ):
        """
        Build query filters and parameters for enterprise analytics queries.

        Arguments:
            enterprise_customer_uuid (UUID): The UUID of the enterprise customer.
            start_date (date): The start date for the query.
            end_date (date): The end date for the query.
            equality_filters (dict, optional): Dictionary of column names and their values to filter by.
                Example: {'course_type': 'OCM', 'is_engaged': 1}
            null_filters (list, optional): List of columns to check for NULL or NOT NULL. Each item should be
                a dict with 'column' and 'null_check' keys.
                Example: [{'column': 'email', 'null_check': True}]
            in_filters (dict, optional): Dictionary of column names and their values to filter by using IN clause.
                Example: {'email': ['user1@example.com', 'user2@example.com']}

        Returns:
            tuple: A tuple containing:
                - QueryFilters: The filters to apply to the query.
                - dict: The parameters to use in the query.
        """
        optional_params = {}
        default_params = {
            'enterprise_customer_uuid': enterprise_customer_uuid,
            'start_date': start_date,
            'end_date': end_date,
        }

        query_filters = QueryFilters([
            self.engagement_filters.enterprise_customer_uuid_filter('enterprise_customer_uuid'),
            self.engagement_filters.date_range_filter(
                column=date_column,
                start_date_params_key='start_date',
                end_date_params_key='end_date',
            ),
        ])

        if equality_filters:
            for column, value in equality_filters.items():
                if value is not None:
                    query_filters.append(EqualQueryFilter(
                        column=column,
                        value_placeholder=column,
                    ))
                    optional_params[column] = value

        if in_filters:
            for column, values_list in in_filters.items():
                params = {f'{column}_{index}': value for index, value in enumerate(values_list)}
                query_filters.append(INQueryFilter(
                    column=column,
                    values=values_list
                ))

        if null_filters:
            for null_column in null_filters:
                query_filters.append(NULLQueryFilter(
                    column=null_column['column'],
                    null_check=null_column['null_check']
                ))

        params = {**default_params, **optional_params}
        return query_filters, params

    @cache_it()
    def _get_engagement_data_for_leaderboard(
            self,
            enterprise_customer_uuid: UUID,
            start_date: date,
            end_date: date,
            limit: int,
            offset: int,
            include_null_email: bool,
            course_type: Optional[str] = None,
            course_key: Optional[str] = None,
            budget_uuid: Optional[str] = None,
            group_uuid: Optional[UUID] = None
    ):
        """
        Get the engagement data for leaderboard.

        The engagement data would include fields like learning time, session length of
        the enterprise learners to show in the leaderboard.

        Arguments:
            enterprise_customer_uuid (UUID): The UUID of the enterprise customer.
            start_date (date): The start date.
            end_date (date): The end date.
            limit (int): The maximum number of records to return.
            offset (int): The number of records to skip.
            include_null_email (bool): If True, only fetch data for NULL emails.
            course_type (Optional[str]): The course type (OCM or Executive Education) to filter by (optional).
            course_key (Optional[str]): The course key to filter by (optional).
            budget_uuid (Optional[str]): The budget UUID to filter by (optional).
            group_uuid (Optional[UUID]): The UUID of a group.

        Returns:
            list[dict]: The engagement data for leaderboard.
        """
        equality_filters = {
            'course_product_line': course_type,
            'course_key': course_key,
            'subsidy_access_policy_uuid': budget_uuid,
            'is_engaged': 1
        }
        query_filters, params = self.build_query_filters_for_leaderboard(
            enterprise_customer_uuid=enterprise_customer_uuid,
            date_column='activity_date',
            start_date=start_date,
            end_date=end_date,
            equality_filters=equality_filters,
        )

        # If group_uuid is provided, we need to filter by enterprise users in the group.
        group_filter_response = self.engagement_filters.enterprise_user_query_filter(
            group_uuid,
            enterprise_customer_uuid
        )
        if group_filter_response is not None:
            enterprise_user_id_in_filter, enterprise_user_id_params = group_filter_response
            query_filters.append(enterprise_user_id_in_filter)
            params.update(enterprise_user_id_params)

        engagements = run_query(
            query=self.queries.get_engagement_data_for_leaderboard_query(query_filters),
            params={
                **params,
                **{
                    'limit': limit,
                    'offset': offset,
                }
            },
            as_dict=True,
        )

        if include_null_email:
            null_filters = [{
                'column': 'email',
                'null_check': True
            }]
            query_filters, params = self.build_query_filters_for_leaderboard(
                enterprise_customer_uuid=enterprise_customer_uuid,
                date_column='activity_date',
                start_date=start_date,
                end_date=end_date,
                equality_filters=equality_filters,
                null_filters=null_filters
            )

            # If group_uuid is provided, we need to filter by enterprise users in the group.
            if group_filter_response is not None:
                enterprise_user_id_in_filter, enterprise_user_id_params = group_filter_response
                query_filters.append(enterprise_user_id_in_filter)
                params.update(enterprise_user_id_params)

            engagement_for_null_email = run_query(
                query=self.queries.get_engagement_data_for_leaderboard_null_email_only_query(query_filters),
                params=params,
                as_dict=True,
            )
            engagements += engagement_for_null_email
        return engagements

    @cache_it()
    def _get_completion_data_for_leaderboard_query(
            self,
            enterprise_customer_uuid: UUID,
            start_date: date,
            end_date: date,
            email_list: list,
            include_null_email: bool,
            course_type: Optional[str] = None,
            course_key: Optional[str] = None,
            budget_uuid: Optional[str] = None,
            group_uuid: Optional[UUID] = None
    ):
        """
        Get the completion data for leaderboard.

        The completion data would include fields like course completion count of enterprise learners of the
        given enterprise customer.

        Arguments:
            enterprise_customer_uuid (UUID): The UUID of the enterprise customer.
            start_date (date): The start date.
            end_date (date): The end date.
            email_list (list<str>): List of emails of the enterprise learners.
            include_null_email (bool): If True, only fetch data for NULL emails.
            course_type (Optional[str]): The course type (OCM or Executive Education) to filter by (optional).
            course_key (Optional[str]): The course key to filter by (optional).
            budget_uuid (Optional[str]): The budget UUID to filter by (optional).
            group_uuid (Optional[UUID]): The UUID of a group.

        Returns:
            list[dict]: The engagement data for leaderboard.
        """
        equality_filters = {
            'course_product_line': course_type,
            'course_key': course_key,
            'subsidy_access_policy_uuid': budget_uuid,
            'has_passed': 1
        }
        in_filters = {
            'email': email_list
        }
        query_filters, params = self.build_query_filters_for_leaderboard(
            enterprise_customer_uuid=enterprise_customer_uuid,
            date_column='passed_date',
            start_date=start_date,
            end_date=end_date,
            equality_filters=equality_filters,
            in_filters=in_filters,
        )

        # If group_uuid is provided, we need to filter by enterprise users in the group.
        group_filter_response = self.engagement_filters.enterprise_user_query_filter(
            group_uuid,
            enterprise_customer_uuid
        )
        if group_filter_response is not None:
            enterprise_user_id_in_filter, enterprise_user_id_params = group_filter_response
            query_filters.append(enterprise_user_id_in_filter)
            params.update(enterprise_user_id_params)

        completions = run_query(
            query=self.queries.get_completion_data_for_leaderboard_query(query_filters),
            params=params,
            as_dict=True,
        )

        if include_null_email:
            query_filters, params = self.build_query_filters_for_leaderboard(
                enterprise_customer_uuid=enterprise_customer_uuid,
                date_column='passed_date',
                start_date=start_date,
                end_date=end_date,
                equality_filters=equality_filters,
                null_filters=[{'column': 'email', 'null_check': True}]
            )

            # If group_uuid is provided, we need to filter by enterprise users in the group.
            if group_filter_response is not None:
                enterprise_user_id_in_filter, enterprise_user_id_params = group_filter_response
                query_filters.append(enterprise_user_id_in_filter)
                params.update(enterprise_user_id_params)

            completions_for_null_email = run_query(
                query=self.queries.get_completion_data_for_leaderboard_null_email_only_query(query_filters),
                params=params,
                as_dict=True,
            )
            completions += completions_for_null_email

        return completions

    def get_all_leaderboard_data(
            self,
            enterprise_customer_uuid: UUID,
            start_date: date,
            end_date: date,
            limit: int,
            offset: int,
            total_count: int,
            course_type: Optional[str] = None,
            course_key: Optional[str] = None,
            budget_uuid: Optional[str] = None,
            group_uuid: Optional[UUID] = None
    ):
        """
        Get the leaderboard data for the given enterprise customer.

        Arguments:
            enterprise_customer_uuid (UUID): The UUID of the enterprise customer.
            start_date (date): The start date.
            end_date (date): The end date.
            limit (int): The maximum number of records to return.
            offset (int): The number of records to skip.
            total_count (int): The total number of records.
            course_type (Optional[str]): The course type filter.
            course_key (Optional[str]): The course key filter.
            budget_uuid (Optional[str]): The budget UUID filter.
            group_uuid (Optional[UUID]): The UUID of a group.

        Returns:
            list[dict]: The leaderboard data.
        """
        include_null_email = False
        # If this is the last or only page, we need to include NULL emails record.
        if total_count <= offset + limit:
            include_null_email = True

        engagement_data = self._get_engagement_data_for_leaderboard(
            enterprise_customer_uuid=enterprise_customer_uuid,
            start_date=start_date,
            end_date=end_date,
            limit=limit,
            offset=offset,
            include_null_email=include_null_email,
            course_type=course_type,
            course_key=course_key,
            budget_uuid=budget_uuid,
            group_uuid=group_uuid,
        )
        # If there is no data, no need to proceed.
        if not engagement_data:
            return []

        engagement_data_dict = {
            engagement['email']: engagement for engagement in engagement_data if engagement['email']
        }
        completion_data = self._get_completion_data_for_leaderboard_query(
            enterprise_customer_uuid=enterprise_customer_uuid,
            start_date=start_date,
            end_date=end_date,
            email_list=list(engagement_data_dict.keys()),
            include_null_email=include_null_email,
            course_type=course_type,
            course_key=course_key,
            budget_uuid=budget_uuid,
            group_uuid=group_uuid,
        )
        for completion in completion_data:
            email = completion['email']
            engagement_data_dict[email]['course_completion_count'] = completion['course_completion_count']

        if include_null_email:
            engagement_data_dict['None'] = find_first(engagement_data, lambda x: x['email'] is None) or {}
            completion = find_first(completion_data, lambda x: x['email'] is None) or \
                {'course_completion_count': ''}
            engagement_data_dict['None']['course_completion_count'] = completion['course_completion_count']
            engagement_data_dict['None']['email'] = NULL_EMAIL_TEXT

        return list(engagement_data_dict.values())

    @cache_it()
    def get_leaderboard_data_count(
        self,
        enterprise_customer_uuid: UUID,
        start_date: date,
        end_date: date,
        course_type: Optional[str] = None,
        course_key: Optional[str] = None,
        budget_uuid: Optional[str] = None,
        group_uuid: Optional[UUID] = None
    ):
        """
        Get the total number of leaderboard records for the given enterprise customer.

        Arguments:
            enterprise_customer_uuid (UUID): The UUID of the enterprise customer.
            start_date (date): The start date.
            end_date (date): The end date.
            course_type (Optional[str]): The course type filter.
            course_key (Optional[str]): The course key filter.
            budget_uuid (Optional[str]): The budget UUID filter.
            group_uuid (Optional[UUID]): The UUID of a group.

        Returns:
            (int): The total number of leaderboard records.
        """
        equality_filters = {
            'course_product_line': course_type,
            'course_key': course_key,
            'subsidy_access_policy_uuid': budget_uuid,
            'is_engaged': 1
        }

        query_filters, params = self.build_query_filters_for_leaderboard(
            enterprise_customer_uuid=enterprise_customer_uuid,
            date_column='activity_date',
            start_date=start_date,
            end_date=end_date,
            equality_filters=equality_filters,
        )

        # If group_uuid is provided, we need to filter by enterprise users in the group.
        group_filter_response = self.engagement_filters.enterprise_user_query_filter(
            group_uuid,
            enterprise_customer_uuid
        )
        if group_filter_response is not None:
            enterprise_user_id_in_filter, enterprise_user_id_params = group_filter_response
            query_filters.append(enterprise_user_id_in_filter)
            params.update(enterprise_user_id_params)

        results = run_query(
            query=self.queries.get_leaderboard_data_count_query(query_filters),
            params=params
        )
        if not results:
            return 0
        return results[0][0]



"""
Constants for admin analytics app.
"""

from enum import Enum


class CourseType(Enum):
    """Course type choices"""
    OCM = "OCM"
    EXECUTIVE_EDUCATION = "Executive Education"


class ResponseType(Enum):
    """Response type choices"""
    JSON = 'json'
    CSV = 'csv'
